
PT_SIZE_4K = 4 * 1024
PT_SIZE_16K = 16 * 1024
PT_SIZE_64K = 64 * 1024
PT_SIZE_1MIB = 1024 * 1024
PT_SIZE_2MIB = 2 * PT_SIZE_1MIB
PT_SIZE_32MIB = 32 * PT_SIZE_1MIB
PT_SIZE_512MIB = 512 * PT_SIZE_1MIB
PT_SIZE_1GIB = 1024 * 1024 * 1024
PT_SIZE_64GIB = 64 * PT_SIZE_1GIB
PT_SIZE_512GIB = 512 * PT_SIZE_1GIB
PT_SIZE_1TB = 1024 * PT_SIZE_1GIB
PT_SIZE_4TB = 4 * PT_SIZE_1TB
PT_SIZE_128TB = 128 * PT_SIZE_1TB



from pt.pt_common import *
from pt.pt_arch_backend import PTArchBackend
from pt.pt_constants import *

import math

class Riscv64_Page():
    def __init__(self, va, phys, size, readable, writeable, executable, user):
        self.va = va
        self.page_size = size
        self.r = readable
        self.w = writeable
        self.x = executable
        self.s = not user
        self.phys = [phys]
        self.sizes = [size]

    def read_memory(self, machine):
        memory = b""
        for phys_range_start, phys_range_size in zip(self.phys, self.sizes):
            memory += machine.read_physical_memory(phys_range_start, phys_range_size)
        return memory

    def cut_before(self, va):
        print("cut_before not supported")

    def cut_after(self, va):
        print("cut_after not supported")

    def pwndbg_is_writeable(self):
        return self.w

    def pwndbg_is_executable(self):
        return self.x

    def to_string(self, phys_verbose):
        prefix = ""
        if not self.s:
            prefix = bcolors.CYAN + " " + bcolors.ENDC
        elif self.s:
            prefix = bcolors.MAGENTA + " " + bcolors.ENDC

        varying_str = None
        if phys_verbose:
            fmt = f"{{:>{PrintConfig.va_len}}} : {{:>{PrintConfig.page_size_len}}} : {{:>{PrintConfig.phys_len}}}"
            varying_str = fmt.format(hex(self.va), hex(self.page_size), hex(self.phys[-1]))
        else:
            fmt = f"{{:>{PrintConfig.va_len}}} : {{:>{PrintConfig.page_size_len}}}"
            varying_str = fmt.format(hex(self.va), hex(self.page_size))
        s = f"{varying_str} | W:{int(self.w)} X:{int(self.x)} R:{int(self.r)} S:{int(self.s)}"

        res = ""
        if self.x and self.w:
            res = prefix + bcolors.BLUE + " " + s + bcolors.ENDC
        elif self.w and not self.x:
            res = prefix + bcolors.GREEN + " " + s + bcolors.ENDC
        elif self.x:
            res = prefix + bcolors.RED + " " + s + bcolors.ENDC
        else:
            res = prefix + " " + s

        return res

class PT_RiscV64_Backend(PTArchBackend):
    def __init__(self, machine):
        self.machine = machine

    def get_arch(self):
        return "riscv64"

    def riscv64_semantically_similar(p1, p2) -> bool:
        return p1.x == p2.x and p1.w == p2.w and p1.r == p2.r and p1.s == p2.s

    def parse_entries(self, table, as_size, lvl):
        dirs = []
        pages = []
        entries = []
        try:
            entries = split_range_into_int_values(read_page(self.machine, table.phys[0]), 8)
        except:
            pass
        for i, pa in enumerate(entries):
            valid = extract(pa, 0, 0)
            if valid:
                is_leaf = (extract(pa, 1, 3) != 0)
                address_contrib = (i << (as_size - lvl * 9))
                child_va = table.va | address_contrib
                phys_addr = extract(pa, 10, 53) << 12
                if is_leaf:
                    size = 1 << (as_size - lvl * 9)
                    readable = extract(pa, 1, 1)
                    writeable = extract(pa, 2, 2)
                    executable = extract(pa, 3, 3)
                    user_accessible = extract(pa, 4, 4)
                    pages.append(Riscv64_Page(child_va, phys_addr, size, readable, writeable, executable, user_accessible))
                else:
                    dirs.append(Riscv64_Page(child_va, phys_addr, None, None, None, None, None))
        return dirs, pages

    def traverse_table(self, pt_addr, as_size):
        root = Riscv64_Page(0, pt_addr, 0, 0, 0, 0, 0)
        directories, leafs = self.parse_entries(root, as_size, lvl=1)

        lvl = 2
        while len(directories) != 0:
            directories_cur_lvl = []
            for tmp_tb in directories:
                tmp_dirs, tmp_leafs = self.parse_entries(tmp_tb, as_size, lvl=lvl)
                directories_cur_lvl.extend(tmp_dirs)
                leafs.extend(tmp_leafs)
            lvl = lvl + 1
            directories = directories_cur_lvl

        for leaf in leafs:
            leaf.va = make_canonical(leaf.va, as_size)

        return leafs

    def print_stats(self):
        raise Exception("Unimplemented")

    def get_address_space_size_from_mode(self, mode_value):
        if mode_value == 8:
            return 39
        elif mode_value == 9:
            return 48
        elif mode_value == 10:
            return 57
        elif mode_value == 11:
            return 64
        else:
            raise Exception(f"Unknown mode: {hex(mode_value)}")

    def walk(self, va):
        entry_size = 8
        num_entries_per_page = int(4096 / entry_size)
        bits_per_level = int(math.log2(num_entries_per_page))

        satp = self.machine.read_register("$satp")
        pt_addr = extract(satp, 0, 43) << 12
        mode_value = extract(satp, 60, 63)
        as_size = get_address_space_size_from_mode(mode_value)

        pt_walk = PageTableWalkInfo(va)
        pt_walk.add_register_stage("satp", pt_addr)

        iter = 0
        while True:
            top_bit = as_size - 1 - iter * bits_per_level
            low_bit = top_bit - bits_per_level + 1
            entry_index = extract(va, low_bit, top_bit)
            entry_page_pa = pt_addr + entry_index * entry_size
            entry_value = int.from_bytes(self.machine.read_physical_memory(entry_page_pa, entry_size), 'little')
            entry_value_pa_no_meta = (extract(entry_value, 10, 53)) << 12
            meta_bits = extract_no_shift(entry_value, 0, 9)
            pt_walk.add_stage(f"Level{iter}", entry_index, entry_value_pa_no_meta, meta_bits)

            if extract(meta_bits, 0, 0) == 0:
                # Not present
                pt_walk.set_faulted()

            is_leaf = (extract(meta_bits, 1, 3) != 0)
            if is_leaf:
                break

            pt_addr = entry_value_pa_no_meta
            iter += 1

        return pt_walk

    def get_filter_is_writeable(self, has_superuser_filter, has_user_filter):
        return lambda p: p.w

    def get_filter_is_not_writeable(self, has_superuser_filter, has_user_filter):
        return lambda p: not p.w

    def get_filter_is_executable(self, has_superuser_filter, has_user_filter):
        return lambda p: p.x

    def get_filter_is_not_executable(self, has_superuser_filter, has_user_filter):
        return lambda p: not p.x

    def get_filter_is_writeable_or_executable(self, has_superuser_filter, has_user_filter):
        return lambda p: p.w or p.x

    def get_filter_is_user_page(self, has_superuser_filter, has_user_filter):
        return lambda p: not p.s

    def get_filter_is_superuser_page(self, has_superuser_filter, has_user_filter):
        return lambda p: p.s

    def get_filter_is_read_only_page(self, has_superuser_filter, has_user_filter):
        return lambda p: p.r and not p.w and not p.x

    def get_filter_architecture_specific(self, filter_name, has_superuser_filter, has_user_filter):
        raise exception(f"Uknown filter {filter_name}")

    def parse_tables(self, cache, args):
        satp = args.satp
        requires_physical_contiguity = args.phys_verbose

        if satp:
            satp = int(satp[0], 16)
        else:
            satp = self.machine.read_register("$satp")

        all_blocks = None

        if satp in cache:
            all_blocks = cache[satp]
        else:
            mode_value = extract(satp, 60, 63)
            as_size = self.get_address_space_size_from_mode(mode_value)
            pt_base = extract(satp, 0, 43) << 12
            all_blocks = self.traverse_table(pt_base, as_size)
            all_blocks = optimize([], [], all_blocks, PT_RiscV64_Backend.riscv64_semantically_similar, requires_physical_contiguity)

        if args.save:
            cache[satp] = all_blocks

        return all_blocks

    def print_kaslr_information(self, table, should_print = True, phys_verbose = False):
        return None

    def print_table(self, table, phys_verbose):
        varying_str = None
        if phys_verbose:
            fmt = f"  {{:>{PrintConfig.va_len}}} : {{:>{PrintConfig.page_size_len}}} : {{:>{PrintConfig.phys_len}}}"
            varying_str = fmt.format("Address", "Length", "Phys")
        else:
            fmt = f"  {{:>{PrintConfig.va_len}}} : {{:>{PrintConfig.page_size_len}}}"
            varying_str = fmt.format("Address", "Length")
        print(bcolors.BLUE + varying_str + "   Permissions    " + bcolors.ENDC)
        for page in table:
            print(page.to_string(phys_verbose))
        return None




from abc import ABC
from abc import abstractmethod

class PTArchBackend(ABC):

    @abstractmethod
    def get_arch(self):
        pass

    @abstractmethod
    def get_filter_is_writeable(self, has_superuser_filter, has_user_filter):
        pass

    @abstractmethod
    def get_filter_is_not_writeable(self, has_superuser_filter, has_user_filter):
        pass

    @abstractmethod
    def get_filter_is_executable(self, has_superuser_filter, has_user_filter):
        pass

    @abstractmethod
    def get_filter_is_not_executable(self, has_superuser_filter, has_user_filter):
        pass

    @abstractmethod
    def get_filter_is_writeable_or_executable(self, has_superuser_filter, has_user_filter):
        pass

    @abstractmethod
    def get_filter_is_user_page(self, has_superuser_filter, has_user_filter):
        pass

    @abstractmethod
    def get_filter_is_superuser_page(self, has_superuser_filter, has_user_filter):
        pass

    @abstractmethod
    def get_filter_is_read_only_page(self, has_superuser_filter, has_user_filter):
        pass

    @abstractmethod
    def get_filter_is_read_only_page(self, has_superuser_filter, has_user_filter):
        pass

    @abstractmethod
    def get_filter_architecture_specific(self, filter_name, has_superuser_filter, has_user_filter):
        pass

    @abstractmethod
    def parse_tables(self, cache, args):
        pass

    @abstractmethod
    def print_table(self, table, phys_verbose):
        pass

    @abstractmethod
    def print_kaslr_information(self, table, phys_verbose):
        pass

    @abstractmethod
    def print_stats(self):
        pass

    @abstractmethod
    def walk(self, va):
        pass




import sys
import argparse
import os
import subprocess
import tempfile
import string
import random
import traceback

from pt.pt_common import *

class PageTableDump():

    def __init__(self, machine_backend, arch_backend, needs_pid = False):
        self.machine_backend = machine_backend
        self.arch_backend = arch_backend

        self.parser = argparse.ArgumentParser()
        self.parser.add_argument("-save", action="store_true")
        self.parser.add_argument("-list", action="store_true")
        self.parser.add_argument("-clear", action="store_true")
        self.parser.add_argument("-ss", nargs='+', type=lambda s: str(s))
        self.parser.add_argument("-sb", nargs='+', type=lambda s: b"".join([int(s[u:u+2], 16).to_bytes(1, 'little') for u in range(0, len(s), 2)]))
        self.parser.add_argument("-s8", nargs='+', type=lambda s: int(s, 0))
        self.parser.add_argument("-s4", nargs='+', type=lambda s: int(s, 0))
        self.parser.add_argument("-range", nargs=2, type=lambda s: int(s, 0))
        self.parser.add_argument("-after", nargs=1, type=lambda s: int(s, 0))
        self.parser.add_argument("-before", nargs=1, type=lambda s: int(s, 0))
        self.parser.add_argument("-has", nargs=1, type=lambda s: int(s, 0))
        self.parser.add_argument("-align", nargs='+', type=lambda s: int(s, 0))
        self.parser.add_argument("-kaslr", action="store_true")
        self.parser.add_argument("-kaslr_leaks", action="store_true")
        self.parser.add_argument("-info", action="store_true")
        self.parser.add_argument("-walk", nargs=1, type=lambda s: int(s, 0))
        self.parser.add_argument("-phys_verbose", action="store_true")
        self.parser.add_argument("-filter", nargs="+")
        self.parser.add_argument("-o", nargs=1)
        self.parser.add_argument("-find_alias", action="store_true")
        self.parser.add_argument("-force_traverse_all", action="store_true")
        self.parser.add_argument("-read_phys", nargs=2, help="hex physical address and length")
        self.parser.add_argument("-read_virt", nargs=2, help="hex virtual address and length")

        if needs_pid:
            self.parser.add_argument("-pid", type=int, required = True)

        if self.arch_backend.get_arch() == "x86_64" or self.arch_backend.get_arch() == "x86_32":
            self.parser.add_argument("-cr3", nargs=1)

        if self.arch_backend.get_arch() == "aarch64":
            self.parser.add_argument("-ttbr0_el1", nargs=1)
            self.parser.add_argument("-ttbr1_el1", nargs=1)

        if self.arch_backend.get_arch() == "riscv64":
            self.parser.add_argument("-satp", nargs=1)

        self.cache = dict()

    def print_cache(self):
        print("Cache:")
        for address in self.cache:
            print(f"\t{hex(address)}")

    def handle_command_wrapper(self, argv):
        args = None
        try:
            args = self.parser.parse_args(argv)
        except:
            return None

        saved_stdout = None
        if args.o:
            saved_stdout = sys.stdout
            sys.stdout = open(args.o[0], "w+")

        try:
            self.handle_command(args)
        except Exception as e:
            print(f"Exception: {str(e)}")
            print(f"Stack trace:\n{traceback.format_exc()}")
        finally:
            if saved_stdout:
                sys.stdout.close()
                sys.stdout = saved_stdout

    def read_virt_memory(self, virt_ranges, virt_addr, len):
        phys_blocks = []
        for r in virt_ranges:
            if r.va + r.page_size <= virt_addr:
                continue
            if r.va >= virt_addr + len:
                break

            acc = 0
            for (phys_addr, phys_extent) in zip(r.phys, r.sizes):
                if r.va + acc >= virt_addr + len:
                    break
                acc += phys_extent
                if r.va + acc < virt_addr:
                    continue
                start_delta = max(virt_addr - (acc + r.va - phys_extent), 0)
                phys_addr_fixed = phys_addr + start_delta
                phys_extent_fixed = min(virt_addr + len + phys_extent - (r.va + acc), phys_extent) - start_delta
                phys_blocks.append((phys_addr_fixed, phys_extent_fixed))

        remaining = len
        data = b""
        for (pa, extent) in phys_blocks:
            to_read = min(remaining, extent)
            data += self.machine_backend.read_physical_memory(pa, to_read)
            remaining -= to_read
            if remaining == 0:
                break
        return data

    def handle_command(self, args):
        if args.list:
            self.print_cache()
            return

        if args.clear:
            self.cache = dict()
            return

        to_search = None
        to_search_num = 0x100000000
        if args.ss:
            to_search = args.ss[0].encode("ascii")
            if len(args.ss) > 1:
                to_search_num = int(args.ss[1], 0)
        if args.sb:
            to_search = args.sb[0]
            if len(args.sb) > 1:
                to_search_num = int.from_bytes(args.sb[1], 'little')
        elif args.s8:
            to_search = args.s8[0].to_bytes(8, 'little')
            if len(args.s8) > 1:
                to_search_num = int(args.s8[1], 0)
        elif args.s4:
            to_search = args.s4[0].to_bytes(4, 'little')
            if len(args.s4) > 1:
                to_search_num = int(args.s4[1], 0)

        requires_page_table_parsing = True
        if args.info:
            requires_page_table_parsing = False

        if args.walk:
            requires_page_table_parsing = False

        if args.read_phys:
            requires_page_table_parsing = False

        page_ranges = None
        page_ranges_filtered = None
        if requires_page_table_parsing:
            page_ranges = self.arch_backend.parse_tables(self.cache, args)
            compound_filter, (min_address, max_address) = self.parse_filter_args(args)
            page_ranges_filtered = list(filter(compound_filter, page_ranges))
            # Perform cut-off of start and end.
            # Only the first and last page entry need to be potentially modified because they were already filtered
            if len(page_ranges_filtered) >= 1:
                if min_address:
                    page_ranges_filtered[0].cut_after(min_address)
                if max_address:
                    page_ranges_filtered[-1].cut_before(max_address)


        if to_search:
            if page_ranges_filtered:
                aligned_to = args.align[0] if args.align else 1
                aligned_offset = args.align[1] if args.align and len(args.align) == 2 else 0
                search_results = search_memory(self.machine_backend, page_ranges_filtered, to_search, to_search_num, aligned_to, aligned_offset)
                for entry in search_results:
                    print("Found at " + hex(entry[0]) + " in " + entry[1].to_string(args.phys_verbose))
            else:
                print("Not found")
        elif args.walk:
            walk = self.arch_backend.walk(args.walk[0])
            print(walk)
        elif args.kaslr:
            self.arch_backend.print_kaslr_information(page_ranges)
        elif args.kaslr_leaks:
            def inner_find_leaks(x, off):
                top = (x >> (off * 8)).to_bytes(8 - off, 'little')
                num_entries = 10
                entries = search_memory(self.machine_backend, page_ranges_filtered, top, num_entries, 1, 0)
                if entries:
                    print(f"Search for {hex(x)}")
                    for entry in entries:
                        print("Found at " + hex(entry[0] - off) + " in " + entry[1].to_string(args.phys_verbose))
            leaks = self.arch_backend.print_kaslr_information(page_ranges, False)
            if leaks:
                inner_find_leaks(leaks[0], 3)
                inner_find_leaks(leaks[1], 5)
        elif args.info:
            self.arch_backend.print_stats()
        elif args.find_alias:
            find_aliases(page_ranges, args.phys_verbose)
        elif args.read_phys:
            phys_addr = int(args.read_phys[0], 0)
            phys_len = int(args.read_phys[1], 0)
            sys.stdout.buffer.write(self.machine_backend.read_physical_memory(phys_addr, phys_len))
        elif args.read_virt:
            virt_addr = int(args.read_virt[0], 0)
            virt_len = int(args.read_virt[1], 0)
            data = self.read_virt_memory(page_ranges, virt_addr, virt_len)
            sys.stdout.buffer.write(data)
        else:
            self.arch_backend.print_table(page_ranges_filtered, phys_verbose=args.phys_verbose)

    def parse_filter_args(self, args):
        filters = []
        min_address = 0
        max_address = 2 ** 64
        if args.range:
            filters.append(lambda page: page.va >= args.range[0] and page.va <= args.range[1])
            min_address = max(args.range[0], min_address)
            max_address = min(args.range[1], max_address)

        if args.has:
            filters.append(lambda page: args.has[0] >= page.va and args.has[0] < page.va + page.page_size)

        if args.after:
            filters.append(lambda page: args.after[0] < page.va + page.page_size)
            min_address = max(args.after[0], min_address)
        else:
            min_address = None

        if args.before:
            filters.append(lambda page: args.before[0] > page.va)
            max_address = min(args.before[0], max_address)
        else:
            max_address = None

        if args.filter:
            # First, we have to determine if user/superuser filter flag was set
            # This is necessary at least for aarch64 where the AP bits provide many possibilities.

            has_superuser_filter = False
            has_user_filter = False
            for f in args.filter:
                if f == "s":
                    has_superuser_filter = True
                if f == "u":
                    has_user_filter = True
            if not has_superuser_filter and not has_user_filter:
                has_superuser_filter = True
                has_user_filter = True
            for f in args.filter:
                if f == "w":
                    filters.append(self.arch_backend.get_filter_is_writeable(has_superuser_filter, has_user_filter))
                elif f == "_w":
                    filters.append(self.arch_backend.get_filter_is_not_writeable(has_superuser_filter, has_user_filter))
                elif f == "x":
                    filters.append(self.arch_backend.get_filter_is_executable(has_superuser_filter, has_user_filter))
                elif f == "_x":
                    filters.append(self.arch_backend.get_filter_is_not_executable(has_superuser_filter, has_user_filter))
                elif f == "w|x" or f == "x|w":
                    filters.append(self.arch_backend.get_filter_is_writeable_or_executable(has_superuser_filter, has_user_filter))
                elif f == "u":
                    filters.append(self.arch_backend.get_filter_is_user_page(has_superuser_filter, has_user_filter))
                elif f == "s":
                    filters.append(self.arch_backend.get_filter_is_superuser_page(has_superuser_filter, has_user_filter))
                elif f == "ro":
                    filters.append(self.arch_backend.get_filter_is_read_only_page(has_superuser_filter, has_user_filter))
                elif f in ["wb", "_wb", "uc", "_uc"]:
                    filters.append(self.arch_backend.get_filter_architecture_specific(f, has_superuser_filter, has_user_filter))
                else:
                    print(f"Unknown filter: {f}")
                    return

        return (create_compound_filter(filters), (min_address, max_address))




from pt.pt_x86_64_definitions import *
from pt.pt_x86_msr import *
from pt.pt_common import *
from pt.pt_constants import *
from pt.pt_arch_backend import PTArchBackend
from abc import ABC
from abc import abstractmethod

import math

class PT_x86_Common_Backend():
 
    def get_filter_is_writeable(self, has_superuser_filter, has_user_filter):
        return lambda p: p.w

    def get_filter_is_not_writeable(self, has_superuser_filter, has_user_filter):
        return lambda p: not p.w

    def get_filter_is_executable(self, has_superuser_filter, has_user_filter):
        return lambda p: p.x

    def get_filter_is_not_executable(self, has_superuser_filter, has_user_filter):
        return lambda p: not p.x

    def get_filter_is_writeable_or_executable(self, has_superuser_filter, has_user_filter):
        return lambda p: p.x or p.w

    def get_filter_is_user_page(self, has_superuser_filter, has_user_filter):
        return lambda p: not p.s

    def get_filter_is_superuser_page(self, has_superuser_filter, has_user_filter):
        return lambda p: p.s

    def get_filter_is_read_only_page(self, has_superuser_filter, has_user_filter):
        return lambda p: not p.x and not p.w

    def get_filter_architecture_specific(self, filter_name, has_superuser_filter, has_user_filter):
        if filter_name == "wb":
            return lambda p: p.wb
        elif filter_name == "_wb":
            return lambda p: not p.wb
        elif filter_name == "uc":
            return lambda p: p.uc
        elif filter_name == "_uc":
            return lambda p: not p.uc
        else:
            return None

    def print_table(self, table, phys_verbose):
        varying_str = None
        if phys_verbose:
            fmt = f"  {{:>{PrintConfig.va_len}}} : {{:>{PrintConfig.page_size_len}}} : {{:>{PrintConfig.phys_len}}} |"
            varying_str = fmt.format("Address", "Length", "Phys")
        else:
            fmt = f"  {{:>{PrintConfig.va_len}}} : {{:>{PrintConfig.page_size_len}}} |"
            varying_str = fmt.format("Address", "Length")
        print(bcolors.BLUE + varying_str + " Permissions          " + bcolors.ENDC)
        for page in table:
            print(page.to_string(phys_verbose))

    def print_stats(self):
        print(self.pt_cr0.check())
        print(self.pt_cr4.check())

    @abstractmethod
    def get_arch(self):
        raise NotImplementedError("")

    def walk(self, va):

        if self.has_paging_enabled() == False:
            raise Exception("Paging is not enabled")

        entry_size = self.get_entry_size()
        num_entries_per_page = int(4096 / entry_size)
        bits_per_level = int(math.log2(num_entries_per_page))

        pse = self.retrieve_pse()
        pse_ignore = self.get_arch() == "x86_64"

        pt_addr = self.machine.read_register("$cr3")

        pt_walk = PageTableWalkInfo(va)
        pt_walk.add_register_stage("CR3", pt_addr)

        top_va = None
        stages = None
        if self.is_long_mode_enabled():
            if self.has_level_5_paging_enabled():
                stages = ["PML5", "PML4", "PDP", "PD", "PT"]
                top_va_bit = 56
            else:
                stages = ["PML4", "PDP", "PD", "PT"]
                top_va_bit = 47
        else:
            stages = ["PD", "PT"]
            top_va_bit = 31

        cur_phys_addr = pt_addr
        for (stage_index, stage_str) in enumerate(stages):
            page_pa = cur_phys_addr & ~0xFFF
            entry_index = extract(va, top_va_bit - bits_per_level + 1, top_va_bit)
            entry_page_pa = page_pa + entry_index * entry_size
            entry_value = int.from_bytes(self.machine.read_physical_memory(entry_page_pa, entry_size), 'little')
            entry_value_pa_no_meta = extract_no_shift(entry_value, 12, 47)
            meta_bits = extract_no_shift(entry_value, 0, 11)

            pt_walk.add_stage(stage_str, entry_index, entry_value_pa_no_meta, meta_bits)

            if not is_present(entry_value):
                pt_walk.set_faulted()
                break

            if is_big_page(entry_value) and (pse or pse_ignore):
                break

            cur_phys_addr = entry_value
            top_va_bit = top_va_bit - bits_per_level

        return pt_walk

    def print_kaslr_information(self, table, should_print=True, phys_verbose=False):
        potential_base_filter = lambda p: p.x and p.s and p.phys[0] % PT_SIZE_2MIB == 0
        tmp = list(filter(potential_base_filter, table))
        found_page = None

        for page in tmp:
            first_byte = self.machine.read_physical_memory(page.phys[0], 1)
            if first_byte[0] == 0x48:
                found_page = page
                break

        stdout_output = ""
        kaslr_addresses = []
        if found_page:
            stdout_output += "Found virtual image base:\n"
            stdout_output += "\tVirt: " + found_page.to_string(phys_verbose) + "\n"
            stdout_output += "\tPhys: " + hex(found_page.phys[0]) + "\n"
            kaslr_addresses.append(found_page.va)
            first_bytes = self.machine.read_physical_memory(page.phys[0], 32)
            page_ranges_subset = filter(lambda page: not page.x and page.s and page.va % PT_SIZE_2MIB == 0, table)
            search_res_iter = search_memory(self.machine, page_ranges_subset, first_bytes, 1, 1, 0)
            try:
                search_res = next(search_res_iter)
                stdout_output += "Found phys map base:\n"
                phys_map_virt_base = search_res[0] - found_page.phys[0]
                phys_map_range = next(range for range in table if range.va >= phys_map_virt_base and phys_map_virt_base < range.va + range.page_size)
                stdout_output += "\tVirt: " + hex(phys_map_virt_base) + " in " + phys_map_range.to_string(phys_verbose) + "\n"
                kaslr_addresses.append(phys_map_virt_base)
            except StopIteration:
                print("Phys map was not found")
        else:
            stdout_output = "Failed to find KASLR info"
        if should_print:
            print(stdout_output)
        return kaslr_addresses

    def retrieve_pse(self):
        return (self.machine.read_register("$cr4") >> 4) & 0x1 == 0x1

    def retrieve_pae(self):
        return (self.machine.read_register("$cr4") >> 5) & 0x1 == 0x1

    def has_paging_enabled(self):
        return (self.machine.read_register("$cr0") >> 31) & 0x1 == 0x1

    def has_level_5_paging_enabled(self):
        return (self.machine.read_register("$cr4") >> 12) & 0x1 == 0x1

    def parse_pml5(self, addr, force_traverse_all):
        entries = []
        entry_size = 8
        try:
            values = split_range_into_int_values(read_page(self.machine, addr), entry_size)
        except:
            return entries
        pml5_cache = {}
        for u, value in enumerate(values):
            if (value & 0x1) != 0: # Page must be present
                if force_traverse_all or value not in pml5_cache:
                    entry = PML5_Entry(value, u)
                    entries.append(entry)
                    pml5_cache[value] = entry
        return entries

    def parse_pml5es(self, pml5es, force_traverse_all, entry_size):
        entries = []
        for pml5e in pml5es:
            pdpe = self.parse_pml4(pml5e, force_traverse_all)
            entries.extend(pdpe)
        return entries

    def parse_pml4(self, pml5e, force_traverse_all):
        entries = []
        entry_size = 8
        try:
            values = split_range_into_int_values(read_page(self.machine, pml5e.pml4), entry_size)
        except:
            return entries
        pml4_cache = {}
        for u, value in enumerate(values):
            if (value & 0x1) != 0: # Page must be present
                if force_traverse_all or value not in pml4_cache:
                    entry = PML4_Entry(value, pml5e.virt_part, u)
                    entries.append(entry)
                    pml4_cache[value] = entry
        return entries

    def parse_pml4es(self, pml4es, force_traverse_all, entry_size):
        entries = []
        for pml4e in pml4es:
            pdpe = self.parse_pdp(pml4e, force_traverse_all, 4096, entry_size)
            entries.extend(pdpe)
        return entries

    def parse_pdp(self, pml4e, force_traverse_all, size, entry_size):
        entries = []
        try:
            values = split_range_into_int_values(self.machine.read_physical_memory(pml4e.pdp, size), entry_size)
        except:
            return entries
        pdp_cache = {}
        for u, value in enumerate(values):
            if (value & 0x1) != 0:
                if force_traverse_all or value not in pdp_cache:
                    entry = PDP_Entry(value, pml4e.virt_part, u)
                    entries.append(entry)
                    pdp_cache[value] = entry
        return entries

    def parse_pdpes(self, pdpes, force_traverse_all, entry_size, pde_shift):
        entries = []
        pages = []
        for pdpe in pdpes:
            if pdpe.large_page == False:
                pdes = self.parse_pd(pdpe, force_traverse_all, entry_size, pde_shift)
                entries.extend(pdes)
            else:
                page = create_page_from_pdpe(pdpe)
                pages.append(page)
        return entries, pages

    def parse_pd(self, pdpe, force_traverse_all, entry_size, pde_shift):
        entries = []
        try:
            values = split_range_into_int_values(read_page(self.machine, pdpe.pd), entry_size)
        except:
            return entries
        pd_cache = {}
        for u, value in enumerate(values):
            if (value & 0x1) != 0:
                if force_traverse_all or value not in pd_cache:
                    entry = PD_Entry(value, pdpe.virt_part, u, pde_shift)
                    entries.append(entry)
                    pd_cache[value] = entry
        return entries

    def parse_pdes(self, pdes, entry_size=8):
        entries = []
        pages = []
        for pde in pdes:
            if pde.big_page == False:
                ptes = self.parse_pt(pde, entry_size)
                entries.extend(ptes)
            else:
                page = create_page_from_pde(pde)
                pages.append(page)
        return entries, pages

    def parse_pt(self, pde, entry_size=8):
        entries = []
        try:
            values = split_range_into_int_values(read_page(self.machine, pde.pt), entry_size)
        except:
            return entries
        for u, value in enumerate(values):
            if (value & 0x1) != 0:
                entry = PT_Entry(value, pde.virt_part, u)
                entries.append(entry)
        return entries

class PT_x86_64_Backend(PT_x86_Common_Backend, PTArchBackend):

    def get_arch(self):
        return "x86_64"

    def __init__(self, machine):
        self.machine = machine
        self.init_registers()

    def init_registers(self):
        self.pt_cr0 = PT_CR0(self.machine)
        self.pt_cr4 = PT_CR4(self.machine)

    def is_long_mode_enabled(self):
        efer = self.machine.read_register("$efer")
        long_mode_enabled = bool((efer >> 8) & 0x1)
        return long_mode_enabled

    def get_entry_size(self):
        if self.is_long_mode_enabled():
            return 8
        else:
            pae = self.retrieve_pae()
            return 8 if pae else 4

    def get_pde_shift(self):
        if self.is_long_mode_enabled():
            return 21
        else:
            pse = self.retrieve_pse()
            pae = self.retrieve_pae()
            if pse and pae:
                # PSE is ignored when PAE is available.
                return 21
            elif not pse and pae:
                # Only PAE. Page size is 2MiB
                return 21
            elif pse and not pae:
                # Only PSE. Page size is 4MiB.
                return 22
            elif not pse and not pae:
                # Default.
                # Manual suggests this shouldn't be possible because the page extension bit in the pde would be ignored.
                # Yet, QEMU doesn't respect this rule and here we are.
                return 21

    def parse_tables(self, cache, args):
        # Check that paging is enabled, otherwise no point to continue.
        if self.has_paging_enabled() == False:
            raise Exception("Paging is not enabled")

        # Check if long mode is enabled
        efer = self.machine.read_register("$efer")
        long_mode_enabled = bool((efer >> 8) & 0x1)

        requires_physical_contiguity = args.phys_verbose
        pt_addr = None
        if args.cr3:
            pt_addr = int(args.cr3[0], 16)
        else:
            pt_addr = self.machine.read_register("$cr3")
            # TODO: Check if these attribute bits in the cr3 need to be respected.
        pt_addr = pt_addr & (~0xfff)

        page_ranges = None

        if pt_addr in cache:
            page_ranges = cache[pt_addr]
        elif long_mode_enabled:
            pde_shift = self.get_pde_shift()
            entry_size = self.get_entry_size()
            pml4es = []
            if self.has_level_5_paging_enabled():
                pml5es = self.parse_pml5(pt_addr, args.force_traverse_all)
                pml4es = self.parse_pml5es(pml5es, args.force_traverse_all, entry_size)
            else:
                pml4es = self.parse_pml4(PML5_Entry(pt_addr, 0), args.force_traverse_all)

            pdpes = self.parse_pml4es(pml4es, args.force_traverse_all, entry_size)
            pdes, large_pages = self.parse_pdpes(pdpes, args.force_traverse_all, entry_size, pde_shift)
            ptes, big_pages = self.parse_pdes(pdes)
            small_pages = []
            for pte in ptes:
                small_pages.append(create_page_from_pte(pte))
            page_ranges = optimize(large_pages, big_pages, small_pages, rwxs_semantically_similar, requires_physical_contiguity)
        else:
            pae = self.retrieve_pae()
            pde_shift = self.get_pde_shift()
            entry_size = self.get_entry_size()

            pdpes = None
            if pae:
                dummy_pml4 = PML4_Entry(pt_addr, 0)
                num_entries = 4
                pdpes = parse_pdp(dummy_pml4, args.force_traverse_all, num_entries * entry_size, entry_size)
            else:
                pdpes = [PDP_Entry(pt_addr, 0, 0)]

            pdes, large_pages = self.parse_pdpes(pdpes, args.force_traverse_all, entry_size, pde_shift)
            ptes, big_pages = self.parse_pdes(pdes, entry_size)
            small_pages = []
            for pte in ptes:
                small_pages.append(create_page_from_pte(pte))
            page_ranges = optimize(large_pages, big_pages, small_pages, rwxs_semantically_similar, requires_physical_contiguity)

        # Cache the page table if caching is set.
        # Caching happens before the filter is applied.
        if args.save:
            cache[pt_addr] = page_ranges

        return page_ranges




from pt.pt_register import *

# Used the `Armv8, for Armv8-A architecture profile` manual.
# I hope this doesn't break any license. Please don't sue :(

class PT_TCR(PT_Register):
    def __init__(self, machine):
        super(PT_TCR, self).__init__(machine, "TCR_EL1", "Translation Control Register EL1 (TCR EL1)")
        self.add_range("T0SZ", 0, 5, lambda x: f"{x} bits are truncated. TTBR0_EL1 addresses {64 - x} bits.")
        self.add_range("EPD0", 7, 7, PT_Decipher_Meaning_Match( \
            {0: "Perform translation table walk using TTBR0_EL1", \
             1: "A TLB miss on an address translated from TTBR0_EL1 generates a Translation fault. No translation table walk is performed."}))
        self.add_range("IRGN0", 8, 9, PT_Decipher_Meaning_Passthrough)
        self.add_range("ORGN0", 10, 11, PT_Decipher_Meaning_Passthrough)
        self.add_range("SH0", 12, 13, PT_Decipher_Meaning_Match( \
            {0b00: "Non-shareable.", \
             0b01: "Reserved.", \
             0b10: "Outer Shareable.", \
             0b11: "Inner Shareable."}))
        self.add_range("TG0", 14, 15, PT_Decipher_Meaning_Match( \
            {0b00: "4 KiB TTBR0_EL1 granule size", \
             0b01: "64 KiB TTBR0_EL1 granule size.", \
             0b10: "16 KiB TTBR0_EL1 granule size."}))
        self.add_range("T1SZ", 16, 21, lambda x: f"{x} bits are truncated. TTBR1_EL1 addresses {64 - x} bits.")
        self.add_range("A1", 22, 22, PT_Decipher_Meaning_Match( \
            {0: "TTBR0_EL1.ASID defines the ASID.", \
             1: "TTBR1_EL1.ASID defines the ASID."}))
        self.add_range("EPD1", 23, 23, PT_Decipher_Meaning_Match( \
            {0: "Perform translation table walk using TTBR1_EL1", \
             1: "A TLB miss on an address translated from TTBR1_EL1 generates a Translation fault. No translation table walk is performed."}))
        self.add_range("IRGN1", 24, 25, PT_Decipher_Meaning_Passthrough)
        self.add_range("ORGN1", 26, 27, PT_Decipher_Meaning_Passthrough)
        self.add_range("SH1", 28, 29, PT_Decipher_Meaning_Match( \
            {0b00: "Non-shareable.", \
             0b01: "Reserved.", \
             0b10: "Outer Shareable.", \
             0b11: "Inner Shareable."}))
        self.add_range("TG1", 30, 31, PT_Decipher_Meaning_Match( \
            {0b01: "16 KiB TTBR1_EL1 granule size", \
             0b10: "4 KiB TTBR1_EL1 granule size.", \
             0b11: "64 KiB TTBR1_EL1 granule size."}))
        self.add_range("IPS", 32, 34, PT_Decipher_Meaning_Match( \
            {0b000: "32 bits, 4 GB.", \
             0b001: "36 bits, 64 GB.", \
             0b010: "40 bits, 1 TB.", \
             0b011: "42 bits, 4 TB.", \
             0b100: "44 bits, 16 TB.", \
             0b101: "48 bits, 256 TB.", \
             0b110: "52 bits, 4 PB."}))
        self.add_range("AS", 36, 36, PT_Decipher_Meaning_Match( \
            {0: "8 bit - the upper 8 bits of TTBR0_EL1 and TTBR1_EL1 are ignored by hardware.", \
             1: "16 bit - the upper 16 bits of TTBR0_EL1 and TTBR1_EL1 are used for allocation and matching in the TLB."}))
        self.add_range("TBI0", 37, 37, PT_Decipher_Meaning_Match( \
            {0: "Top Byte used in the address calculation.",
             1: "Top Byte ignored in the address calculation."}))
        self.add_range("TBI1", 38, 38, PT_Decipher_Meaning_Match( \
            {0: "Top Byte used in the address calculation.",
             1: "Top Byte ignored in the address calculation."}))
 
